/***
  This file is part of eudev, forked from systemd.

  Copyright 2010 Lennart Poettering

  systemd is free software; you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  systemd is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with systemd; If not, see <http://www.gnu.org/licenses/>.
***/

#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>

#include "macro.h"
#include "util.h"
#include "missing.h"
#include "log.h"
#include "strv.h"
#include "path-util.h"
#include "hashmap.h"
#include "conf-files.h"

static int files_add(Hashmap *h, const char *root, const char *path, const char *suffix) {
        _cleanup_closedir_ DIR *dir = NULL;
        char *dirpath;

        assert(path);
        assert(suffix);

        dirpath = strjoina(root ? root : "", path);

        dir = opendir(dirpath);
        if (!dir) {
                if (errno == ENOENT)
                        return 0;
                return -errno;
        }

        for (;;) {
                struct dirent *de;
                char *p;
                int r;

                errno = 0;
                de = readdir(dir);
                if (!de && errno != 0)
                        return -errno;

                if (!de)
                        break;

                if (!dirent_is_file_with_suffix(de, suffix))
                        continue;

                p = strjoin(dirpath, "/", de->d_name, NULL);
                if (!p)
                        return -ENOMEM;

                r = hashmap_put(h, eudev_basename(p), p);
                if (r == -EEXIST) {
                        log_debug("Skipping overridden file: %s.", p);
                        free(p);
                } else if (r < 0) {
                        free(p);
                        return r;
                } else if (r == 0) {
                        log_debug("Duplicate file %s", p);
                        free(p);
                }
        }

        return 0;
}

static int base_cmp(const void *a, const void *b) {
        const char *s1, *s2;

        s1 = *(char * const *)a;
        s2 = *(char * const *)b;
        return strcmp(eudev_basename(s1), eudev_basename(s2));
}

static int conf_files_list_strv_internal(char ***strv, const char *suffix, const char *root, char **dirs) {
        _cleanup_hashmap_free_ Hashmap *fh = NULL;
        char **files, **p;
        int r;

        assert(strv);
        assert(suffix);

        /* This alters the dirs string array */
        if (!path_strv_resolve_uniq(dirs, root))
                return -ENOMEM;

        fh = hashmap_new(&string_hash_ops);
        if (!fh)
                return -ENOMEM;

        STRV_FOREACH(p, dirs) {
                r = files_add(fh, root, *p, suffix);
                if (r == -ENOMEM) {
                        return r;
                } else if (r < 0)
                        log_debug("Failed to search for files in %s: %s",
                                  *p, strerror(-r));
        }

        files = hashmap_get_strv(fh);
        if (files == NULL) {
                return -ENOMEM;
        }

        qsort_safe(files, hashmap_size(fh), sizeof(char *), base_cmp);
        *strv = files;

        return 0;
}

int conf_files_list_strv(char ***strv, const char *suffix, const char *root, const char* const* dirs) {
        _cleanup_strv_free_ char **copy = NULL;

        assert(strv);
        assert(suffix);

        copy = strv_copy((char**) dirs);
        if (!copy)
                return -ENOMEM;

        return conf_files_list_strv_internal(strv, suffix, root, copy);
}

/* Copy every non-empty path element from path_variable to
   destination, and add a final NULL pointer. If destination is NULL,
   then only count the number of non-empty path elements. Return
   -ENOMEM if the copy failed, or the number of elements. */
static ssize_t conf_files_list_follow_path (const char *path_variable, char **destination) {
        /* This function is not using strtok_r, because two passes are
           required: the first one to count the number of elements,
           the second one to actually copy them once an array of the
           correct size has been allocated. */
        size_t n = 0;
        if (path_variable != NULL) {
                while (path_variable[0] == ':') {
                      path_variable++;
                }
                while (path_variable[0] != '\0') {
                        const char *end = strchr(path_variable, ':');
                        if (end == NULL) {
                                end = path_variable + strlen(path_variable);
                        }
                        if (destination != NULL) {
                                destination[n] = strndup(path_variable, end - path_variable);
                                if (destination[n] == NULL) {
                                        return -ENOMEM;
                                }
                        }
                        n++;
                        path_variable = end;
                        while (path_variable[0] == ':') {
                                path_variable++;
                        }
                }
        }
        if (destination != NULL) {
                destination[n] = NULL;
        }
        return n;
}

/* path is a colon-separated list of directories. */
int conf_files_list_strv_path(char ***strv, const char *suffix, const char *root, const char* path) {
        _cleanup_strv_free_ char **copy = NULL;

        assert(strv);
        assert(suffix);

        ssize_t path_length = conf_files_list_follow_path(path, NULL);
        if (path_length < 0)
                return -ENOMEM;

        copy = new0(char *, path_length + 1);
        if (!copy)
                return -ENOMEM;
        ssize_t error = conf_files_list_follow_path(path, copy);
        if (error < 0)
                return -ENOMEM;

        assert (error == path_length);

        return conf_files_list_strv_internal(strv, suffix, root, copy);
}
